
/*

  charset for CPS1 implementation

*/


#include "cps1charset.h"

//#define DEBUG

cps1charset::cps1charset(int nspr,int nchr,int nt16,int nt32)
{
	long int c;

	gfxrom=0;
	nSprites=nspr;
	nChars=nchr;
	nTiles16=nt16;
	nTiles32=nt32;

	// allocate graphic charsets memory space

	sprites=new unsigned char* [nSprites];
	chars=new unsigned char* [nChars];
	tiles16=new unsigned char* [nTiles16];
	tiles32=new unsigned char* [nTiles32];

	for (c=0;c<nSprites;c++)
	{
		sprites[c]=new unsigned char [16*16];
		memset(sprites[c],0,16*16);
	}
	for (c=0;c<nChars;c++)
	{
		chars[c]=new unsigned char [8*8];
		memset(chars[c],0,8*8);
	}
	for (c=0;c<nTiles16;c++)
	{
		tiles16[c]=new unsigned char [16*16];
		memset(tiles16[c],0,16*16);
	}
	for (c=0;c<nTiles32;c++)
	{
		tiles32[c]=new unsigned char [32*32];
		memset(tiles32[c],0,32*32);
	}

	//

	tile16opaque=new char [nTiles16];
}

cps1charset::~cps1charset()
{
	// free decoded charsets

	if (sprites!=0)
	{
		for (int j=0;j<nSprites;j++)
		{
			if (sprites[j]!=0) delete [] sprites[j];
		}

		delete [] sprites;
	}

	if (chars!=0)
	{
		for (int j=0;j<nChars;j++)
		{
			if (chars[j]!=0) delete [] chars[j];
		}

		delete [] chars;
	}

	if (tiles16!=0)
	{
		for (int j=0;j<nTiles16;j++)
		{
			if (tiles16[j]!=0) delete [] tiles16[j];
		}

		delete [] tiles16;
	}

	if (tiles32!=0)
	{
		for (int j=0;j<nTiles32;j++)
		{
			if (tiles32[j]!=0) delete [] tiles32[j];
		}

		delete [] tiles32;
	}

	delete [] tile16opaque;
}

void cps1charset::printSprite(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy)
{
	if ( (xp<-16) || (xp>=384) || (yp<-16) || (yp>=224) ) return;

	unsigned long int* pbmp=&vb[yp*384+xp];
	unsigned char* pchar=sprites[charnum];
	unsigned long int* ppal=(unsigned long int*)palette;
	int x,y;
	int c;

	color*=16;

	// have we to do clipping?
	if ( (xp>=0) && (xp<(384-16)) && (yp>=0) && (yp<(224-16)) )
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if (c!=15)
					{
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#else
						*pbmp=ppal[c+color];
#endif
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-16);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=(15*16);
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					//c=pchar[x+(15-y)*16];
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-16);
				pchar-=32;
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=15;
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					//c=pchar[(15-x)+y*16];
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
				pchar+=32;
			}
		}
		else
		{
			pchar+=(15*16)+15;
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					//c=pchar[(15-x)+(15-y)*16];
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
			}
		}
	}
	else
	// clip
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#else
						*pbmp=ppal[c+color];
#endif
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-16);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=15*16;
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-16);
				pchar-=32;
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=15;
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
				pchar+=32;
			}
		}
		else
		{
			pchar+=(15*16+15);
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
			}
		}
	}
}

void cps1charset::printChar(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy)
{
	unsigned long int* pbmp=&vb[yp*384+xp];
	unsigned char* pchar=chars[charnum];
	int x,y;
	int c;

	unsigned long int* pcpal=(unsigned long int*)&palette[0x200*4];
	color*=16;

	if ((!flipx) && (!flipy))
	{
		for (y=0;y<8;y++)
		{
			if ((yp+y)>=224) return;
			for (x=0;x<8;x++)
			{
				c=*pchar;

				if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
				else if (c!=15)
				{
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#else
					*pbmp=pcpal[c+color];
#endif
				}
				pbmp++;
				pchar++;
			}
			pbmp+=(384-8);
		}
	}
	else if ((!flipx) && (flipy))
	{
		pchar+=7*8;
		for (y=0;y<8;y++)
		{
			if ((yp+y)>=224) return;
			for (x=0;x<8;x++)
			{
				c=*pchar;

				if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
				else if (c!=15)
				{
					*pbmp=pcpal[c+color];
				}
				pbmp++;
				pchar++;
			}
			pbmp+=(384-8);
			pchar-=16;
		}
	}
	else if ((flipx) && (!flipy))
	{
		pchar+=7;
		for (y=0;y<8;y++)
		{
			if ((yp+y)>=224) return;
			for (x=0;x<8;x++)
			{
				c=*pchar;

				if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
				else if (c!=15)
				{
					*pbmp=pcpal[c+color];
				}
				pbmp++;
				pchar--;
			}
			pbmp+=(384-8);
			pchar+=16;
		}
	}
	else
	{
		pchar+=(7*8+7);
		for (y=0;y<8;y++)
		{
			if ((yp+y)>=224) return;
			for (x=0;x<8;x++)
			{
				c=*pchar;

				if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
				else if (c!=15)
				{
					*pbmp=pcpal[c+color];
				}
				pbmp++;
				pchar--;
			}
			pbmp+=(384-8);
		}
	}
}

void cps1charset::printTile16(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy)
{
	if ( (xp<-15) || (xp>=384) || (yp<-15) || (yp>=224) ) return;

	unsigned long int* pbmp=&vb[yp*384+xp];
	unsigned char* pchar=tiles16[charnum];
	int x,y;
	int c;

	unsigned long int* ptpal=(unsigned long int*)&palette[0x400*4];
	color*=16;

	// have we to do clipping?
	if ( (xp>=0) && (xp<(384-16)) && (yp>=0) && (yp<(224-16)) )
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#endif
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-16);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=(15*16);
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pchar-=32;
				pbmp+=(384-16);
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=15;
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
				pchar+=32;
			}
		}
		else
		{
			pchar+=(15*16)+15;
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
			}
		}
	}
	else
	// clip blit
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-16);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=(15*16);
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pchar-=32;
				pbmp+=(384-16);
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=15;
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
				pchar+=32;
			}
		}
		else
		{
			pchar+=(15*16)+15;
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
			}
		}
	}
}

//

void cps1charset::printTile16rs(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy,int ofz)
{
	if ( (yp<-15) || (yp>223) ) return;

	unsigned char* pchar=tiles16[charnum];
	unsigned long int* vbp;
	int x,y;
	int c;
	int rso;
	unsigned long int* ptpal=(unsigned long int*)&palette[0x400*4];
	int* rs=rowscroll;

	color*=16;

	if ((!flipx) && (!flipy))
	{
		for (y=0;y<16;y++)
		{
			int ypp=yp+y;
			if (ypp>=224) return;
			rso=rs[ypp+ofz];
			int xpp=xp+rso;

			if (ypp>=0)
			{
				vbp=&vb[xpp+ypp*384];
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( (xpp<0) || (xpp>=384) ) ;
					else if (c!=15)
					{
						*vbp=ptpal[c+color];
					}
					pchar++;
					xpp++;
					vbp++;
				}
			}
			else
			{
				pchar+=16;
			}
		}
	}
	else if ((!flipx) && (flipy))
	{
		pchar+=(15*16);
		for (y=0;y<16;y++)
		{
			int ypp=yp+y;
			if (ypp>=224) return;
			rso=rs[ypp+ofz];
			int xpp=xp+rso;

			if (ypp>=0)
			{
				vbp=&vb[xpp+ypp*384];
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( (xpp<0) || (xpp>=384) ) ;
					else if (c!=15)
					{
						*vbp=ptpal[c+color];
					}
					pchar++;
					xpp++;
					vbp++;
				}
				pchar-=32;
			}
			else
			{
				pchar-=16;
			}
		}
	}
	else if ((flipx) && (!flipy))
	{
		pchar+=15;
		for (y=0;y<16;y++)
		{
			int ypp=yp+y;
			if (ypp>=224) return;
			rso=rs[ypp+ofz];
			int xpp=xp+rso;

			if (ypp>=0)
			{
				vbp=&vb[xpp+ypp*384];
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( (xpp<0) || (xpp>=384) ) ;
					else if (c!=15)
					{
						*vbp=ptpal[c+color];
					}
					pchar--;
					xpp++;
					vbp++;
				}
			}
			else
			{
				pchar-=16;
			}
			pchar+=32;
		}
	}
	else
	{
		pchar+=(15*16)+15;
		for (y=0;y<16;y++)
		{
			int ypp=yp+y;
			if (ypp>=224) return;
			rso=rs[ypp+ofz];
			int xpp=xp+rso;

			if (ypp>=0)
			{
				vbp=&vb[xpp+ypp*384];
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( (xpp<0) || (xpp>=384) ) ;
					else if (c!=15)
					{
						*vbp=ptpal[c+color];
					}
					pchar--;
					xpp++;
					vbp++;
				}
			}
			else
			{
				pchar-=16;
			}
		}
	}
}



//

void cps1charset::printTile32(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy)
{
	if ( (xp<-31) || (xp>=384) || (yp<-31) || (yp>=224) ) return;

	unsigned long int* pbmp=&vb[yp*384+xp];
	unsigned char* pchar=tiles32[charnum];
	unsigned long int* ptpal=(unsigned long int*)&palette[0x600*4];
	int x,y;
	int c;
	color*=16;

	// have we to do clipping?
	if ( (xp>=0) && (xp<(384-32)) && (yp>=0) && (yp<(224-32)) )
	{
		//noclip
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<32;y++)
			{
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
#ifdef DEBUG
						*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#endif
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-32);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=(31*32);
			for (y=0;y<32;y++)
			{
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-32);
				pchar-=64;
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=31;
			for (y=0;y<32;y++)
			{
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-32);
				pchar+=64;
			}
		}
		else
		{
			pchar+=(31*32)+31;
			for (y=0;y<32;y++)
			{
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-32);
			}
		}
	// clip
	}
	else
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<32;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-32);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=(31*32);
			for (y=0;y<32;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-32);
				pchar-=64;
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=31;
			for (y=0;y<32;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-32);
				pchar+=64;
			}
		}
		else
		{
			pchar+=(31*32)+31;
			for (y=0;y<32;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-32);
			}
		}
	}
}
