
/*

  the cps1 romspace roms reading

*/


#include "cps1romspace.h"
#include "cps1charset.h"
#include "gui.h"
#include "roms.h"

extern tGui* gui;
extern tRomSet romsetList[];


// constructor

cps1romspace::cps1romspace()
{
	// zero point roms

	romCPU1=0;
	romCPU2=0;

	romadpcm=0;
}

// destructor

cps1romspace::~cps1romspace()
{
	if (romCPU1!=0) delete [] romCPU1;
	if (romCPU2!=0) delete [] romCPU2;

	if (romadpcm!=0) delete [] romadpcm;
}

// read roms

int cps1romspace::readRoms(char* path,char* romsetname,tCharset** chs)
{
	int err;
	int rs;
	long int i,j;
	unsigned char* temprom;
	char fullname[1024];
	tRomSet* rsp;

	err=romsetByName(romsetname,&rs);
	if (err!=0)
	{
		return 1;
	}

	rsp=&romsetList[rs];

	//
	// read main cpu roms
	// (68000)
	//

	temprom=new unsigned char [0x400000];
	romCPU1=new unsigned char [0x800000];

	memset(romCPU1,0,0x800000);

	for (i=0;i<rsp->ncoderoms;i++)
	{
		// load rom into temprom

		strcpy(fullname,path);
		strcat(fullname,rsp->romlist[i].romset);
		strcat(fullname,".zip");

		err=loadZippedRom(fullname,
						  rsp->romlist[i].name,
						  rsp->romlist[i].size,
						  rsp->romlist[i].crc32,
						  temprom);
		if (err!=0)
		{
			gui->printMessage(". Unable to load code rom");
			return 1;
		}

		// type of rom?

		if (rsp->romlist[i].loadtype==ROM_LOAD_ODD)
		{
			for (j=0;j<rsp->romlist[i].size;j++)
			{
				romCPU1[rsp->romlist[i].position+j*2+1]=temprom[j];
			}
		}
		else if (rsp->romlist[i].loadtype==ROM_LOAD_EVEN)
		{
			for (j=0;j<rsp->romlist[i].size;j++)
			{
				romCPU1[rsp->romlist[i].position+j*2]=temprom[j];
			}
		}
		else if (rsp->romlist[i].loadtype==ROM_LOAD_WIDESWAP)
		{
 			memcpy(romCPU1+rsp->romlist[i].position,
				   temprom,
				   rsp->romlist[i].size);			
		}
		else
		{
			gui->printMessage(".Invalid rom load format");
			return 1;
		}
	}

	//
	// read the graphic roms
	//

	if (*chs!=0) delete *chs;
	*chs=new cps1charset(rsp->nspr,rsp->nchars,rsp->ntile16,rsp->ntile32);

	// calculate and store 
	// total size of graphic roms

	tCharset* pchs=*chs;

	gromsize=getGromSize(rsp);

	pchs->gfxrom=new unsigned char [gromsize];

	for (i=rsp->ncoderoms;i<(rsp->ncoderoms+rsp->ngfxroms);i++)
	{
		strcpy(fullname,path);
		strcat(fullname,rsp->romlist[i].romset);
		strcat(fullname,".zip");

		if (rsp->romlist[i].loadtype==ROM_LOAD_NORMAL)
		{
			err=loadZippedRom(fullname,
							  rsp->romlist[i].name,
							  rsp->romlist[i].size,
							  rsp->romlist[i].crc32,
							  pchs->gfxrom+rsp->romlist[i].position);
			if (err!=0)
			{
				gui->printMessage(".Unable to load graphic rom");
				return 1;
			}
		}
		else if (rsp->romlist[i].loadtype==ROM_LOAD_SKIP6)
		{
			err=loadZippedRom(fullname,
							  rsp->romlist[i].name,
							  rsp->romlist[i].size,
							  rsp->romlist[i].crc32,
							  temprom);
			if (err!=0)
			{
				gui->printMessage(".Unable to load graphic rom");
				return 1;
			}

			// load words skipping 6 bytes
			
			for (int p=0;p<(rsp->romlist[i].size);p+=2)
			{
				pchs->gfxrom[rsp->romlist[i].position+(p/2)*8]=  temprom[p];
				pchs->gfxrom[rsp->romlist[i].position+(p/2)*8+1]=temprom[p+1];
			}
		}
	}

	//
	// read sound roms
	//

	// z80

	romCPU2=new unsigned char [0x20000];
	memset(romCPU2,0,0x20000);

	if (rsp->nsndcoderoms>0)
	{
		i=rsp->ncoderoms+rsp->ngfxroms;

		strcpy(fullname,path);
		strcat(fullname,rsp->romlist[i].romset);
		strcat(fullname,".zip");

		err=loadZippedRom(fullname,
						  rsp->romlist[i].name,
						  rsp->romlist[i].size,
						  rsp->romlist[i].crc32,
						  romCPU2+rsp->romlist[i].position);
		if (err!=0)
		{
			gui->printMessage(".Unable to load sound rom");
			return 1;
		}
	}

	// adpcm data

	romadpcm=new unsigned char [0x40000];

	if (rsp->nsndsampleroms>0)
	{
		i++;

		strcpy(fullname,path);
		strcat(fullname,rsp->romlist[i].romset);
		strcat(fullname,".zip");

		err=loadZippedRom(fullname,
						  rsp->romlist[i].name,
						  rsp->romlist[i].size,
						  rsp->romlist[i].crc32,
						  romadpcm+rsp->romlist[i].position);
		if (err!=0)
		{
			gui->printMessage(".Unable to load 1st adpcm rom");
			return 1;
		}


		i++;

		strcpy(fullname,path);
		strcat(fullname,rsp->romlist[i].romset);
		strcat(fullname,".zip");

		err=loadZippedRom(fullname,
						  rsp->romlist[i].name,
						  rsp->romlist[i].size,
						  rsp->romlist[i].crc32,
						  romadpcm+rsp->romlist[i].position);
		if (err!=0)
		{
			gui->printMessage(".Unable to load 2nd adpcm rom");
			return 1;
		}
	}

	// store emulation related data

	strcpy(romname,rsp->name);
	kludge=rsp->kludge;
	orientation=rsp->orientation;
	romsettype=rsp->romsettype;
	nspr=rsp->nspr;
	nchars=rsp->nchars;
	ntile16=rsp->ntile16;
	ntile32=rsp->ntile32;

	sprmagic1=rsp->sprmagic1; sprmagic2=rsp->sprmagic2;
	charmagic1=rsp->charmagic1;
	tile16magic1=rsp->tile16magic1; tile16magic2=rsp->tile16magic2; 
	tile32magic1=rsp->tile32magic1; tile32magic2=rsp->tile32magic2; 

	sproffs=rsp->sproffs;
	charoffs=rsp->charoffs;
	tile16offs=rsp->tile16offs;
	tile32offs=rsp->tile32offs;

	charbase=rsp->charbase;
	tile16base=rsp->tile16base;
	tile32base=rsp->tile32base;

	charpad=rsp->charpad;
	tile16pad=rsp->tile16pad;
	tile32pad=rsp->tile32pad;

	protport=rsp->protport;
	protval=rsp->protval;
	multlo=rsp->multlo;
	multhi=rsp->multhi;
	fac1=rsp->fac1;
	fac2=rsp->fac2;
	layercontrol=rsp->layercontrol;

	// free space

	delete [] temprom;

	return 0;
}

/*

  functions to decode the charset

*/

void cps1romspace::decodesprites(int num,unsigned char *src,unsigned char** dest)
{
/*SPRITE_LAYOUT(ffight, 4096*2+512, 0x80000*8, 0x100000*8)*/

	int plane,x,y;
	unsigned char* dp;
	unsigned char* tmpdp;
	int nplanes=4;
	int charincrement=32*8;

	const long int f2=sprmagic1;
	const long int f4=sprmagic2;

	long int planeoffset[4]={ f2+8,f2+0, 8, 0 };

	long int xoffset[16]=	{ f4+0,f4+1,f4+2,f4+3,f4+4,f4+5,f4+6,f4+7,
							  0, 1, 2, 3, 4, 5, 6, 7 };
	long int yoffset[16]=	        { 0*8, 2*8, 4*8, 6*8, 8*8, 10*8, 12*8, 14*8,
		           16*8, 18*8, 20*8, 22*8, 24*8, 26*8, 28*8, 30*8, };


	for (plane = 0;plane < nplanes;plane++)
	{
		int offs;

		offs = num*charincrement+planeoffset[plane];

		for (y = 0;y < 16;y++)
		{
			//dp = gfx->gfxdata->line[num * 16 + y];
			//dp=dest[num]+(y*16);
			tmpdp=dest[num];
			dp=tmpdp+y*16;

			for (x = 0;x < 16;x++)
			{
				int xoffs,yoffs;

				if (plane == 0) dp[x] = 0;
				else dp[x] <<= 1;

				xoffs = x;
				yoffs = y;

				dp[x] += readbit(src,offs + yoffset[yoffs] + xoffset[xoffs]);
			}
		}
	}
}

//

void cps1romspace::decodechars(int num,unsigned char *src,unsigned char** dest)
{
	int plane,x,y;
	unsigned char* dp;
	unsigned char* tmpdp;
	int nplanes=4;
	int charincrement=16*8;

	const long int f2=charmagic1;

	long int planeoffset[4]={ f2+8,f2+0, 8, 0 };
	long int xoffset[8]={0,1,2,3,4,5,6,7};
	long int yoffset[8]={ 0*8, 2*8, 4*8, 6*8, 8*8, 10*8, 12*8, 14*8};

	unsigned char* psrc=src+charoffs;

	for (plane = 0;plane < nplanes;plane++)
	{
		int offs;

		offs = num*charincrement+planeoffset[plane];

		for (y = 0;y < 8;y++)
		{
			tmpdp=dest[num];
			dp=tmpdp+y*8;

			for (x = 0;x < 8;x++)
			{
				int xoffs,yoffs;

				if (plane == 0) dp[x] = 0;
				else dp[x] <<= 1;

				xoffs = x;
				yoffs = y;

				dp[x] += readbit(psrc,offs + yoffset[yoffs] + xoffset[xoffs]);
			}
		}
	}
}

//

void cps1romspace::decodetiles16(int num,unsigned char *src,unsigned char** dest)
{
	int plane,x,y;
	unsigned char* dp;
	unsigned char* tmpdp;
	int nplanes=4;
	int charincrement=32*8;

	const long int f2=tile16magic1;
	const long int f4=tile16magic2;

	long int planeoffset[4]={ f2+8,f2+0, 8, 0 };

	long int xoffset[16]=	{ f4+0,f4+1,f4+2,f4+3,f4+4,f4+5,f4+6,f4+7,
							  0, 1, 2, 3, 4, 5, 6, 7 };
	long int yoffset[16]=	        { 0*8, 2*8, 4*8, 6*8, 8*8, 10*8, 12*8, 14*8,
		           16*8, 18*8, 20*8, 22*8, 24*8, 26*8, 28*8, 30*8, };

	unsigned char* psrc=src+tile16offs;

	for (plane = 0;plane < nplanes;plane++)
	{
		int offs;

		offs = num*charincrement+planeoffset[plane];

		for (y = 0;y < 16;y++)
		{
			tmpdp=dest[num];
			dp=tmpdp+y*16;

			for (x = 0;x < 16;x++)
			{
				int xoffs,yoffs;

				if (plane == 0) dp[x] = 0;
				else dp[x] <<= 1;

				xoffs = x;
				yoffs = y;

				dp[x] += readbit(psrc,offs + yoffset[yoffs] + xoffset[xoffs]);
			}
		}
	}
}


//

/*
TILE32_LAYOUT(tile32layout_ffight, 512+128,     0x080000*8, 0x100000*8)
*/

void cps1romspace::decodetiles32(int num,unsigned char *src,unsigned char** dest)
{
	int plane,x,y;
	unsigned char* dp;
	unsigned char* tmpdp;
	int nplanes=4;
	int charincrement=32*8*4;

	const long int f2=tile32magic1;
	const long int SEP=tile32magic2;

	long int planeoffset[4]={ f2+8,f2+0, 8, 0 };

	long int xoffset[32]={                                                   
					SEP+0,SEP+1,SEP+2,SEP+3, SEP+4,SEP+5,SEP+6,SEP+7,
					0,1,2,3,4,5,6,7,                                 
					16+SEP+0,16+SEP+1,16+SEP+2,                      
					16+SEP+3,16+SEP+4,16+SEP+5,                      
					16+SEP+6,16+SEP+7,                               
					16+0,16+1,16+2,16+3,16+4,16+5,16+6,16+7          
					};                                               

	long int yoffset[32]=
		{                                                          
           0*32, 1*32, 2*32, 3*32, 4*32, 5*32, 6*32, 7*32,         
           8*32, 9*32, 10*32, 11*32, 12*32, 13*32, 14*32, 15*32,   
           16*32, 17*32, 18*32, 19*32, 20*32, 21*32, 22*32, 23*32, 
           24*32, 25*32, 26*32, 27*32, 28*32, 29*32, 30*32, 31*32  
        };

	unsigned char* psrc=src+tile32offs;

	for (plane = 0;plane < nplanes;plane++)
	{
		int offs;

		offs = num*charincrement+planeoffset[plane];

		for (y = 0;y < 32;y++)
		{
			tmpdp=dest[num];
			dp=tmpdp+y*32;

			for (x = 0;x < 32;x++)
			{
				int xoffs,yoffs;

				if (plane == 0) dp[x] = 0;
				else dp[x] <<= 1;

				xoffs = x;
				yoffs = y;

				dp[x] += readbit(psrc,offs + yoffset[yoffs] + xoffset[xoffs]);
			}
		}
	}
}

void cps1romspace::decodeCharset(tCharset* chset)
{
	int c;

	for (c = 0;c < chset->nSprites;c++)
	{
		decodesprites(c,chset->gfxrom,chset->sprites);
	}

	for (c = 0;c < chset->nChars;c++)
	{
		decodechars(c,chset->gfxrom,chset->chars);
	}

	for (c = 0;c < chset->nTiles16;c++)
	{
		decodetiles16(c,chset->gfxrom,chset->tiles16);
	}

	for (c = 0;c < chset->nTiles32;c++)
	{
		decodetiles32(c,chset->gfxrom,chset->tiles32);
	}

	// search for opaque tiles
/*
	int nopq=0;

	for (c=0;c<chset->nTiles32;c++)
	{
		int isopaque=1;

		for (int pix=0;pix<32*32;pix++)
		{
			if (chset->tiles32[c][pix]==15)
			{
				isopaque=0;
			}
		}

		if (isopaque) nopq++;
		chset->tile16opaque[c]=isopaque;
	}

	char str[256];
	sprintf(str,"[%d] opaque tiles32 on [%d] total tiles\n",nopq,chset->nTiles32);
	MessageBox(0,str,"info",MB_OK);
*/
}