
/*

  cpu main definitions

*/


#ifndef CPUCORE_H
#define CPUCORE_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

#include "a68k/a68k.h"
#include "makez80/mz80.h"
#include "romspace.h"
#include "sound.h"


// general functions

__inline int READ_WORD(void *dst)
{
	return *((unsigned short int*)dst);
}

__inline int READ8(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	return mem[(addr^1)-base];
}

__inline int READ16(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	unsigned short int* pmem=(unsigned short int*)&mem[addr-base];
	return (*pmem);
}

__inline unsigned long int READ32(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	unsigned short int* pmem=(unsigned short int*)&mem[addr-base];
	return (pmem[0]<<16)|pmem[1];
}

__inline void WRITE8(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[(addr^1)-base]=(unsigned char)val;
}

__inline void WRITE16(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	unsigned short int* pmem=(unsigned short int*)&mem[addr-base];
	*pmem=val;
}

__inline void WRITE32(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	unsigned long int* pmem=(unsigned long int*)&mem[addr-base];
	val=(val>>16)|(val<<16);
	*pmem=val;
}

// cps1 memory functions

unsigned char cps1_m68kr8(int address);
unsigned short int cps1_m68kr16(int address);
unsigned long int cps1_m68kr32(int address);
void cps1_m68kw8(int address, unsigned char value);
void cps1_m68kw16(int address, unsigned short int value);
void cps1_m68kw32(int address, unsigned long int value);

// cps2 memory functions

unsigned short int cps2_m68kri16(int address);
unsigned long int cps2_m68kri32(int address);
unsigned char cps2_m68krpc8(int address);
unsigned short int cps2_m68krpc16(int address);
unsigned long int cps2_m68krpc32(int address);
unsigned char cps2_m68kr8(int address);
unsigned short int cps2_m68kr16(int address);
unsigned long int cps2_m68kr32(int address);
void cps2_m68kw8(int address, unsigned char value);
void cps2_m68kw16(int address, unsigned short int value);
void cps2_m68kw32(int address, unsigned long int value);

// neogeo memory functions

unsigned char neogeo_m68kr8(int address);
unsigned short int neogeo_m68kr16(int address);
unsigned long int neogeo_m68kr32(int address);
void neogeo_m68kw8(int address, unsigned char value);
void neogeo_m68kw16(int address, unsigned short int value);
void neogeo_m68kw32(int address, unsigned long int value);

// the class

class cpucore
{
public:

	cpucore();
	cpucore(tRomSpace*);
	virtual ~cpucore();

	virtual void emulate(int);
	virtual void drawScreen(unsigned long int* video,tCharset* chset);
};

#endif