
/*

  a.c.e.
  Another Capcom (Play System) Emulator
  (aka "l'emulatore candeggina")

  project started 12/5/2001

*/


#include <stdio.h>
#include <stdlib.h>

#include "videowin/videowin.h"
#include "render.h"
#include "sound.h"
#include "charset.h"
#include "cpucore.h"
#include "timing.h"

//
 
tGui* gui;
video* vid;
render* rend;

//

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG msg;

	// init gui

	gui=new tGui(0,384,224,"ace.ini");
	if (gui->initfailed) 
	{
		MessageBox(NULL,"can't start ace","error",MB_OK);
		return 1;
	}
	gui->printMessage(". Ace started");

	// init display

	vid=new videowin();
	vid->setViewport(384,224);
	((videowin*)vid)->setHinstance(hInstance);

	if (vid->init(&gui->videoBuf,gui->guiinfo.videomode,VIDEO_NORMAL))
	{
		MessageBox(NULL,"can't do video init","error",MB_OK);
		return 1;
	}

	// init renderer

	rend=new render(384,224);

	// display greetings

	gui->printMessage(". Init display ok");
	gui->printMessage(". Press right mouse button");

	// init timer
	
	init_timer();

	// init sound

	if (gui->guiinfo.soundon)
	{
		if (initSound(((videowin*)vid)->getHwnd(),gui->guiinfo.soundfreq))
		{
			MessageBox(NULL,"can't do sound init","error",MB_OK);
			vid->free(&gui->videoBuf);
			shutdownSound();
			delete gui;
			return 1;
		}
	}

	// main loop

	while (!gui->exitprogram())
	{
		if (gui->emulating)
		{
			if (!gui->ginput.paused) gui->machine->emulate(gui->guiinfo.soundon);
			
			if ((gui->frames%gui->guiinfo.frameskip)==0) 
			{
				gui->machine->drawScreen(gui->videoBuf,gui->chset);
				rend->apply(gui->videoBuf);
			}

			gui->update();
			gui->redraw(gui->videoBuf);

			if ((gui->frames%gui->guiinfo.frameskip)==0) 
			{
				vid->blit(gui->videoBuf);
				vid->clear(gui->videoBuf);
			}

			gui->frames++;
			if (gui->guiinfo.trimspeed!=0) trim_speed();
			update_fps(gui->frames,gui->onscreenfps,gui->guiinfo.frameskip);
		}
		else
		{
			gui->update();
			gui->redraw(gui->videoBuf);
			vid->blit(gui->videoBuf);
			vid->clear(gui->videoBuf);
			if (gui->guiinfo.trimspeed!=0) trim_speed();
		}

		while (PeekMessage(&msg, NULL,0,0,PM_NOREMOVE))
		{
			GetMessage (&msg, NULL, 0, 0);
			TranslateMessage (&msg);
			DispatchMessage (&msg);
		}
	}

	if (gui->guiinfo.showfps)
	{
		// show fps
		int secs;
		char strsecs[256];
		secs=secs_elapsed();
		sprintf(strsecs,"[fps: %f]",(float)((float)gui->frames/(float)secs));
		MessageBox(0,strsecs,"info",MB_OK);
	}

	// dispose soundz

	if (gui->guiinfo.soundon)
	{
		shutdownSound();
	}

	// dispose video & renderer

	vid->free(&gui->videoBuf);
	delete rend;

	// dispose gui

	delete gui;

	// exit

	return (msg.wParam);
}
