
/*

  NeoGeo charset

  sprite zooming code is by David Raingeard
  author of Calice emulator
  Thank you very much!

  (I have left his comments in french
   because they are great!)

*/


#include "neogeocharset.h"

//#define DEBUG
#define xpad ((384-304)/2)

neogeocharset::neogeocharset(int nspr,int nchr,int nt16,int nt32)
{
	int i;

	gfx1=0;
	gfx2=0;
	nSprites=nspr;
	nChars=nchr;
	sprites=0;

	chars=new unsigned char* [nchr];

	for (i=0;i<nchr;i++)
	{
		chars[i]=new unsigned char [8*8];
		memset(chars[i],0,8*8);
	}
}

neogeocharset::~neogeocharset()
{
	int i;

	if (gfx1!=0) delete [] gfx1;
	if (gfx2!=0) delete [] gfx2;

	for (i=0;i<nChars;i++)
	{
		delete [] chars[i];
	}

	delete [] chars;
}

void neogeocharset::printChar(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy)
{
	if ( (xp<-7+xpad) || (xp>=(384-xpad)) || 
		 (yp<-7) || (yp>=224) ) return;

	unsigned long int* pbmp=&vb[yp*384+xp];
	unsigned char* pchar=chars[charnum];
	int x,y;
	int c;

	color*=16;

	unsigned long int* pcpal=(unsigned long int*)palette;

	if ( (xp>=0+xpad) && 
		 (xp<(384-xpad-8)) && 
		 (yp>=0) && 
		 (yp<(224-8)) )
	{
		for (y=0;y<8;y++)
		{
			for (x=0;x<8;x++)
			{
				c=*pchar;

				if (c!=0)
				{
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#else
					*pbmp=pcpal[c+color];
#endif
				}
				pbmp++;
				pchar++;
			}
			pbmp+=(384-8);
		}
	}
	else
	{
		for (y=0;y<8;y++)
		{
			if ((yp+y)>=224) return;
			for (x=0;x<8;x++)
			{
				c=*pchar;

				if ( ((xp+x)<0+xpad) || ((xp+x)>=(384-xpad)) || ((yp+y)<0) ) ;
				else if (c!=0)
				{
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#else
					*pbmp=pcpal[c+color];
#endif
				}
				pbmp++;
				pchar++;
			}
			pbmp+=(384-8);
		}
	}
}

void neogeocharset::printSprite(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy,
							  int zoomx,int zoomy)
{
	if ( (xp<-16+xpad) || (xp>=384-xpad) || (yp<-16) || (yp>=224) ) return;

	if ( (zoomx==16) && (zoomy==16) )
	{
		/* no scaling */
		unsigned long int* pbmp=&vb[yp*384+xp];
		unsigned char* pchar=&gfx2[charnum*16*8];
		unsigned long int* ppal=(unsigned long int*)palette;
		int x,y;
		int c,cp;

		color*=16;

		// have we to do clipping?
		if ( (xp>=xpad) && (xp<(384-16-xpad)) && (yp>=0) && (yp<(224-16)) )
		{
			if ((!flipx) && (!flipy))
			{
				for (y=0;y<16;y++)
				{
	//				for (x=0;x<16;x++)
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp&0x0f;

						if (c!=0)
						{
	#ifdef DEBUG
						*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
	#else
							*pbmp=ppal[c+color];
	#endif
						}
						pbmp++;

						c=cp>>4;

						if (c!=0)
						{
	#ifdef DEBUG
						*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
	#else
							*pbmp=ppal[c+color];
	#endif
						}
						pbmp++;

						pchar++;
					}
					pbmp+=(384-16);
				}
			}
			else if ((!flipx) && (flipy))
			{
	//			pchar+=(15*16);
				pchar+=(15*8);
				for (y=0;y<16;y++)
				{
	//				for (x=0;x<16;x++)
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp&0x0f;

						if (c!=0)
						{
							*pbmp=ppal[c+color];
						}
						pbmp++;

						c=cp>>4;

						if (c!=0)
						{
							*pbmp=ppal[c+color];
						}
						pbmp++;

						pchar++;
					}
					pbmp+=(384-16);
					pchar-=16;
				}
			}
			else if ((flipx) && (!flipy))
			{
				pchar+=7;
				for (y=0;y<16;y++)
				{
	//				for (x=0;x<16;x++)
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp>>4;

						if (c!=0)
						{
							*pbmp=ppal[c+color];
						}
						pbmp++;

						c=cp&0x0f;

						if (c!=0)
						{
							*pbmp=ppal[c+color];
						}
						pbmp++;
						
						pchar--;
					}
					pbmp+=(384-16);
					pchar+=16;
				}
			}
			else
			{
	//			pchar+=(15*16)+15;
				pchar+=(15*8)+7;
				for (y=0;y<16;y++)
				{
	//				for (x=0;x<16;x++)
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp>>4;

						if (c!=0)
						{
							*pbmp=ppal[c+color];
						}
						pbmp++;

						c=cp&0x0f;

						if (c!=0)
						{
							*pbmp=ppal[c+color];
						}
						pbmp++;

						pchar--;
					}
					pbmp+=(384-16);
				}
			}
		}
		else
		// clip
		{
			if ((!flipx) && (!flipy))
			{
				for (y=0;y<16;y++)
				{
					if ((yp+y)>=224) return;
	//				for (x=0;x<16;x++)
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp&0x0f;

						if ( ((xp+x)>=xpad) && ((xp+x)<384-xpad) && ((yp+y)>=0) && (c!=0) )
						{
	#ifdef DEBUG
						*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
	#else
							*pbmp=ppal[c+color];
	#endif
						}
						pbmp++;

						c=cp>>4;

						if ( ((xp+x)>=-1+xpad) && ((xp+x)<383-xpad) && ((yp+y)>=0) && (c!=0) )
						{
	#ifdef DEBUG
						*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
	#else
							*pbmp=ppal[c+color];
	#endif
						}
						pbmp++;


						pchar++;
					}
					pbmp+=(384-16);
				}
			}
			else if ((!flipx) && (flipy))
			{
	//			pchar+=15*16;
				pchar+=15*8;
				for (y=0;y<16;y++)
				{
					if ((yp+y)>=224) return;
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp&0x0f;

						if ( ((xp+x)>=xpad) && ((xp+x)<384-xpad) && ((yp+y)>=0) && (c!=0) )
						{
							*pbmp=ppal[c+color];
						}
						pbmp++;

						c=cp>>4;

						if ( ((xp+x)>=-1+xpad) && ((xp+x)<383-xpad) && ((yp+y)>=0) && (c!=0) )
						{
							*pbmp=ppal[c+color];
						}
						pbmp++;
						
						pchar++;
					}
					pbmp+=(384-16);
					pchar-=16;
				}
			}
			else if ((flipx) && (!flipy))
			{
				pchar+=7;
				for (y=0;y<16;y++)
				{
					if ((yp+y)>=224) return;
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp>>4;

						if ( ((xp+x)>=xpad) && ((xp+x)<384-xpad) && ((yp+y)>=0) && (c!=0) )
						{
							*pbmp=ppal[c+color];
						}
						pbmp++;

						c=cp&0x0f;

						if ( ((xp+x)>=-1+xpad) && ((xp+x)<383-xpad) && ((yp+y)>=0) && (c!=0) )
						{
							*pbmp=ppal[c+color];
						}
						pbmp++;
						
						pchar--;
					}
					pbmp+=(384-16);
					pchar+=16;
				}
			}
			else
			{
				pchar+=(15*8+7);
				for (y=0;y<16;y++)
				{
					if ((yp+y)>=224) return;
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp>>4;

						if ( ((xp+x)>=xpad) && ((xp+x)<384-xpad) && ((yp+y)>=0) && (c!=0) )
						{
							*pbmp=ppal[c+color];
						}
						pbmp++;

						c=cp&0x0f;

						if ( ((xp+x)>=-1+xpad) && ((xp+x)<383-xpad) && ((yp+y)>=0) && (c!=0) )
						{
							*pbmp=ppal[c+color];
						}
						pbmp++;

						pchar--;
					}
					pbmp+=(384-16);
				}
			}
		}
	}
	else
	{
		unsigned long int* ppal=(unsigned long int*)palette;
		int real_width, real_height;
		unsigned long int *dst;
		int bclx, offset, bcly;
		unsigned int inc_x;
		int fixedWidthIncrement, fixedHeightIncrement;
		unsigned int fixedWidthPos, fixedHeightPos, fixedWidthPosStart;

		if ( (zoomx==0) || (zoomy==0) ) return;

		// unroll the sprite (bleah)

		unsigned char sprite[16*16];
		unsigned char* pchar=&gfx2[charnum*16*8];
		unsigned char* spr_ptr=sprite;

		for (int sy=0;sy<16;sy++)
		{
			for (int sx=0;sx<16;sx+=2)
			{
				sprite[sy*16+sx+1]=((*pchar)>>4);
				sprite[sy*16+sx]=((*pchar)&0x0f);
				pchar++;
			}
		}

		color*=16;

		fixedWidthIncrement  = (16<<16)/zoomx;
		fixedHeightIncrement = (16<<16)/zoomy;

		real_width=zoomx;
		real_height=zoomy;
		inc_x=0;

		// si le sprite dpasse sur la droite
		if ((xp+zoomx)>(319+xpad)) real_width=(319+xpad-xp);  

		// si le sprite dpasse sur le bas
		if ((yp+zoomy)>239) zoomy=(239-yp);

		// si le sprite dpasse sur la gauche
		if (xp<0) 
		{ 
			// 
			real_width+=xp; 
			spr_ptr+=((-xp)*fixedWidthIncrement)>>16;
			xp=0; 
		}

		// si le sprite dpasse sur le bas
		if (yp<0) 
		{ 
			real_height+=yp; 
			spr_ptr+=16*(((-yp)*fixedHeightIncrement)>>16);
			yp=0;
		}

		if ( (real_width<0) || (real_height<0) ) return;

		dst=vb+xp+yp*384;

		// prise en compte des permutations verticales et horizontales
		if (flipx)
		{
			fixedWidthIncrement=-fixedWidthIncrement;
			fixedWidthPosStart=(-fixedWidthIncrement)*(real_width-1);
		}
		else fixedWidthPosStart=0;

		if (flipy)
		{
			fixedHeightPos=fixedHeightIncrement*(real_height-1);
			fixedHeightIncrement=-fixedHeightIncrement;
		}
		else fixedHeightPos=0;

		int mxp=xp;
		int myp=yp;
		int c;

		for (bcly=0;bcly<real_height;bcly++)
		{
			mxp=xp;
			fixedWidthPos=fixedWidthPosStart;
			offset=((fixedHeightPos>>16)<<4);
			for (bclx=0;bclx<real_width;bclx++)
			{
				c=spr_ptr[(offset+(fixedWidthPos>>16))&0xff];
				if (c) 
				{
					if ( (mxp>=xpad) && (mxp<384-xpad) )
					{
						*dst=ppal[c+color];
					}
				}

				dst++;
				mxp++;
				fixedWidthPos+=fixedWidthIncrement;
			}	
			fixedHeightPos+=fixedHeightIncrement;
			dst+=(384-real_width);
			myp++;
		} 
	}
}
